/*
 *  Interactive disassembler (IDA).
 *  NEC V830 module
 */

#include "v830.hpp"

//----------------------------------------------------------------------
static char abcond[] = { 
  V830_ABV,  V830_ABL,  V830_ABZ,  V830_ABNH,
  V830_ABN,  V830_ABR,  V830_ABLT, V830_ABLE, 
  V830_ABNV, V830_ABNC, V830_ABNZ, V830_ABH,  
  V830_ABP,  V830_NOP, V830_ABGE, V830_ABGT};
 
static char bcond[] = {
  V830_BV,  V830_BL,  V830_BZ,  V830_BNH,
  V830_BN,  V830_BR,  V830_BLT, V830_BLE, 
  V830_BNV, V830_BNC, V830_BNZ, V830_BH,  
  V830_BP,  V830_NOP, V830_BGE, V830_BGT};

#define FORMAT5_REG1(x) ((x) & 0x1f)
#define FORMAT5_REG2(x) (((x) >> 5) & 0x1f)
#define OPCODE(x) ((x) >> 10)

int find_movhi(int reg) {
  int i, newea = prevaddr(cmd.ea);
  for (i=0; i<20; i++) {
    int hword = get_word(newea);
    if (OPCODE(hword) == 0x2f) {
      if (FORMAT5_REG1(hword) == rR0 && FORMAT5_REG2(hword) == reg) return newea;
    }
    newea = prevaddr(newea);
  }
  return BADADDR;
}

/* Look back in disassembly for a ld.w statement, such as:
   ld.w    8C3BFB8h(r28), r28
   (the fact that the offset is a dword means we already went back and patched
   the instruction -- however, we don't have access to its cmd structure,
   so we must recognize the byte pattern which would become the above, 
   then run ua_ana0, then proceed.)
   Returns the immediate offset (i.e. 8C3BFB8h)  */

int find_ldw(int reg) {
  insn_t saved = cmd;
  int i, newea = prevaddr(cmd.ea), retval;
  for(i=0; i<20; i++) {
    int hword = get_word(newea);
    if(OPCODE(hword) == 0x33) {
//  msg("Found ld.w at %x\n", newea);
      if (FORMAT5_REG1(hword) == reg && FORMAT5_REG2(hword) == reg) {
//  msg("Passed reg test\n");
	if(ua_ana0(newea) != 4) {
	  msg("ERROR: find_ldw unable to parse ld.w at %x\n", newea);
	  cmd = saved;
	  return BADADDR;
	}
//  msg("Got offset of %x\n", cmd.Op1.addr);
	retval = cmd.Op1.addr;
	cmd = saved;
	return retval;
      }
    }
    newea = prevaddr(newea);
  }
  cmd = saved;
  return BADADDR;
}

void postprocess(void) {
  unsigned int movhi_addr=0, new_offset=0, jmp_target=0;
  // msg("Opcode @ %x: %x\n", cmd.ea, opcode);
  switch(cmd.itype) {
  case V830_MOVEA: // movea
    if (cmd.Op2.reg != cmd.Op3.reg) break;
    movhi_addr = find_movhi(cmd.Op2.reg);
    if (movhi_addr != BADADDR) {
      new_offset = cmd.Op1.value & 0xffff;
      cmd.Op1.type = o_near;
      cmd.Op1.dtyp = dt_dword;
      cmd.Op1.addr = new_offset | (((unsigned int)get_word(movhi_addr+2)) << 16);
      if(new_offset & 0x8000) cmd.Op1.addr -= 0x10000;
    }
    /*if (cmd.Op2.reg == rR31) { // special case for lr -- always a  code ref.
      ea_t target = cmd.Op1.value;
      cmd.Op1.type = o_near;
      cmd.Op1.addr = target;
      } */
    break;
  case V830_LD_B: // ld.b
  case V830_LD_H: // ld.h
  case V830_LD_W: // ld.w
    movhi_addr = find_movhi(cmd.Op1.phrase);
//  msg("find_movhi=%x\n", movhi_addr);
    if (movhi_addr != BADADDR) {
      new_offset = cmd.Op1.addr & 0xffff;
//    msg("new_offset=%x\n", new_offset);
      cmd.Op1.type = o_near;
      cmd.Op1.dtyp = dt_dword;
      cmd.Op1.addr = new_offset | (((unsigned int)get_word(movhi_addr+2)) << 16);
      if(new_offset & 0x8000) cmd.Op1.addr -= 0x10000;
    }
    break;
  case V830_JMP:
    jmp_target = find_ldw(cmd.Op1.reg);
//  msg("find_ldw(%d)=%x\n", cmd.Op1.reg, jmp_target);
    if (jmp_target != BADADDR && (get_long(jmp_target) != 0)) {
      cmd.Op1.addr = get_long(jmp_target);
      cmd.Op1.dtyp = dt_dword;
      cmd.Op1.type = o_near;
    }
    break;
  }
}

//----------------------------------------------------------------------
int __stdcall ana( void ) {
  cmd.Op1.dtyp = dt_byte;
  cmd.Op2.dtyp = dt_byte;

  uint hibyte = ua_next_byte();
  uint lobyte = ua_next_byte();
  uint op_hword = hibyte + (lobyte << 8);
  uint op_word = 0;
  char format=0;
  if (op_hword == 0) return 0;
  uint opcode = (op_hword >> 10) % 64;

  char offc;
  uint tmp=0;
  int stmp=0; // signed
  
  switch(opcode) {
  case 0:  format=1; cmd.itype=V830_MOV; break;
  case 16: format=2; cmd.itype=V830_MOV; break;
  case 1:  format=1; cmd.itype=V830_ADD; break;
  case 17: format=2; cmd.itype=V830_ADD; break;
  case 2:  format=1; cmd.itype=V830_SUB; break;
  case 3:  format=1; cmd.itype=V830_CMP; break;
  case 19: format=2; cmd.itype=V830_CMP; break;
  case 4:  format=1; cmd.itype=V830_SHL; break;
  case 20: format=2; cmd.itype=V830_SHL; break;
  case 5:  format=1; cmd.itype=V830_SHR; break;
  case 21: format=2; cmd.itype=V830_SHR; break;
  case 6:  format=1; cmd.itype=V830_JMP; break;
  case 7:  format=1; cmd.itype=V830_SAR; break;
  case 23: format=2; cmd.itype=V830_SAR; break; 
  case 8:  format=1; cmd.itype=V830_MUL; break;
  case 9:  format=1; cmd.itype=V830_DIV; break;
  case 10: format=1; cmd.itype=V830_MULU; break;
  case 11: format=1; cmd.itype=V830_DIVU; break;
  case 12: format=1; cmd.itype=V830_OR; break;
  case 13: format=1; cmd.itype=V830_AND; break;
  case 14: format=1; cmd.itype=V830_XOR; break;
  case 15: format=1; cmd.itype=V830_NOT; break;
  case 18: format=2; cmd.itype=V830_SETF; break;
  case 22: format=2; cmd.itype=V830_EI; break;
  case 24: format=2; cmd.itype=V830_TRAP; break;
  case 25: format=9; cmd.itype=(op_hword&1?V830_BRKRET:V830_RETI); break;
  case 26: format=9; cmd.itype=(op_hword&1?V830_STBY:V830_HALT); break;
  case 27: return 0;
  case 28: format=2; cmd.itype=V830_LDSR; break;
  case 29: return 0;
  case 30: format=2; cmd.itype=V830_DI; break;
  case 31: return 0;
  case 32: case 33: case 34: case 35:
  case 36: case 37: case 38: case 39:
    format=3;
    cmd.itype = (op_hword&1)?(abcond[(op_hword>>9)&0xF]) : (bcond[(op_hword>>9)&0xF]);
    break;
  case 40: format=5; cmd.itype=V830_MOVEA; break;
  case 41: format=5; cmd.itype=V830_ADDI; break;
  case 42: format=4; cmd.itype=V830_JR; break;
  case 43: format=4; cmd.itype=V830_JAL; break;
  case 44: format=5; cmd.itype=V830_ORI; break;
  case 45: format=5; cmd.itype=V830_ANDI; break;
  case 46: format=5; cmd.itype=V830_XORI; break;
  case 47: format=5; cmd.itype=V830_MOVHI; break;
  case 48: format=6; cmd.itype=V830_LD_B; break;
  case 49: format=6; cmd.itype=V830_LD_H; break;
  case 51: format=6; cmd.itype=V830_LD_W; break;
  case 50: format=5; cmd.itype=V830_MULI; break;
  case 52: format=6; cmd.itype=V830_ST_B; break;
  case 53: format=6; cmd.itype=V830_ST_H; break;
  case 54: format=5; cmd.itype=V830_MACI; break;
  case 55: format=6; cmd.itype=V830_ST_W; break;
  case 56: format=6; cmd.itype=V830_IN_B; break;
  case 57: format=6; cmd.itype=V830_IN_H; break;
  case 59: format=6; cmd.itype=V830_IN_W; break;
  case 58: format=6; cmd.itype=V830_CAXI; break;
  case 60: format=6; cmd.itype=V830_OUT_B; break;
  case 61: format=6; cmd.itype=V830_OUT_B; break;
  case 62: format=7;  break; // "special"
  case 63: format=6; cmd.itype=V830_OUT_B; break;
  default: cmd.itype=V830_null;
    msg("Unknown opcode %d (0x%x)", opcode, opcode);
  }
  
  switch(format) {
  case 0: msg("Unknown format!"); break;
  case 1:   // opcode: 15-10 / reg2: 9-5 / reg1: 4-0 
    cmd.Op1.type = o_reg;
    cmd.Op1.reg = (op_hword & 0x1f) + rR0;
    
    cmd.Op2.type = o_reg;
    cmd.Op2.reg = ((op_hword>>5)  & 0x1f) + rR0;
    
    if(cmd.itype == V830_JMP) {
      cmd.Op2.type=o_void;
      //cmd.Op1.type=o_phrase;
    }
    if(cmd.itype == V830_LDSR || cmd.itype == V830_STSR) {
      cmd.Op1.reg += 32;
    }
    break;
  case 2:   // opcode: 15-10 / reg2: 9-5 / imm5: 4-0
    offc = (op_hword & 0x1f);
    if (op_hword & 0x10) offc |= 0xF0;
    cmd.Op1.type = o_imm;
    cmd.Op1.value = offc;
    
    cmd.Op2.type = o_reg;
    cmd.Op2.reg = ((op_hword>>5)  & 0x1f) + rR0;
    break;
  case 3:   // opcode: 15-13 / cond: 12-9 / disp9: 8-1 / sub-opcode: 0
    offc=char((op_hword>>1) & 0xff);
    cmd.Op1.addr = cmd.ea + (offc*2);  // signed addition
    cmd.Op1.type = o_near;
    cmd.Op1.dtyp = dt_dword;
    break;
  case 4: // opcode: 15-10 / disp25: 9-0, 31-16
    stmp=(op_hword & 0x3ff) << 22;
    stmp /= 64;
    stmp |= ua_next_byte();
    stmp |= ua_next_byte() << 8;
    cmd.Op1.addr = cmd.ip + stmp;
    cmd.Op1.dtyp = dt_dword;
    cmd.Op1.type = o_near;
    break;
  case 5: // opcode: 15-10 / reg2: 9-5 / reg1: 4-0 / imm16: 31-16
    tmp = ua_next_byte();
    tmp |= ua_next_byte() << 8;
    //		msg("cmd.Op1.value = %x\n", cmd.Op1.value);
    cmd.Op1.value = tmp;
    cmd.Op1.dtyp = dt_dword;
    cmd.Op1.type = o_imm;
    //  op_offset(cmd.ea, 0, REF_OFF32, tmp, tmp & 0xFFFF0000);
    
    cmd.Op2.type = o_reg;
    cmd.Op2.reg = (op_hword & 0x1f) + rR0;
    cmd.Op3.type = o_reg;
    cmd.Op3.reg = ((op_hword>>5) & 0x1f) + rR0;
    //			if (cmd.itype==V830_MOVEA) 
    //				msg("ana.cpp@%x: %x\n", cmd.ea, cmd.Op1.value);
    break;
  case 6: // opcode: 15-10 / disp25: 9-0, 31-16
    stmp = ua_next_byte();
    stmp |= ua_next_byte() << 8;
    cmd.Op1.addr = (short)stmp;
    cmd.Op1.phrase = op_hword&0x1f + rR0;
    cmd.Op1.type = o_displ;
    cmd.Op1.dtyp = dt_word;
    cmd.Op2.type = o_reg;
    cmd.Op2.reg = (op_hword>>5)&0x1f + rR0;
    if(cmd.itype==V830_ST_H || cmd.itype == V830_ST_W || cmd.itype == V830_ST_B) {
      cmd.Op3=cmd.Op1;
      cmd.Op1=cmd.Op2;
      cmd.Op2=cmd.Op3;
      cmd.Op3.type=o_void;
    }
    //msg("op3=%d, op1=%d, stmp=%hd", op_hword&0x1f, (op_hword>>6)&0x1f, stmp);
    break;
  case 7: // opcode: 15-10 / disp25: 9-0, 31-16
    op_word = ua_next_byte() << 16;
    op_word |= ua_next_byte() << 24;
    op_word |= op_hword;
    switch(op_word >> 26) {
    case 0x10: cmd.itype = V830_SATADD3; break; 
    case 0x11: cmd.itype = V830_SATSUB3; break;
    case 0x12: cmd.itype = V830_MIN3; break;
    case 0x13: cmd.itype = V830_MAX3; break;
    case 0x18: cmd.itype = V830_SHLD3; break;
    case 0x19: cmd.itype = V830_SHRD3; break;
    case 0x1c: cmd.itype = V830_MACT3; break;
    case 0x1d: cmd.itype = V830_MAC3; break;
    case 0x1e: cmd.itype = V830_MULT3; break;
    case 0x1f: cmd.itype = V830_MUL3; break;
    case 0x20: cmd.itype = V830_BILD; break;
    case 0x21: cmd.itype = V830_BDLD; break;
    case 0x22: cmd.itype = V830_BIST; break;
    case 0x23: cmd.itype = V830_BDST; break;
    default:
      msg("%x:Unknown special opcode %x (num=%x)\n", cmd.ea, op_word, op_word >> 26);
      return 0;
    }
    cmd.Op1.type = o_reg;
    cmd.Op1.reg = op_word & 0x1f;
    cmd.Op2.type = o_reg;
    cmd.Op2.reg = (op_word >> 5) & 0x1f;
    if((op_word >> 26) < 0x20) {
      cmd.Op3.type = o_reg;
      cmd.Op3.reg = (op_word >> 16) & 0x1f;
    }
    break;
  case 8: // opcode: 15-10 / disp25: 9-0, 31-16
    stmp=(op_hword & 0x1ff) << 23;
    stmp /= 64;
    stmp |= ua_next_byte();
    stmp |= ua_next_byte() << 8;
    cmd.Op1.addr = cmd.ip  + stmp;
    cmd.Op1.dtyp = dt_word;
    cmd.Op1.type = o_near;
    break;
  case 9: break;
  default:
    msg("unknown format %d\n", format);
  }
  
  postprocess();
  // msg("returning cmd.itype = %d format = %d\n", cmd.itype, format);
  return cmd.size;
}
