/*
 *  Interactive disassembler (IDA).
 *  NEC V830 module
 *
 */

#include <ida.hpp>
#include <idp.hpp>
#include "ins.hpp"

instruc_t Instructions[] = {
  { "",       0                       },  // Unknown Operation

/* load/store */
  { "ld.b",       CF_USE1|CF_USE2|CF_USE3|CF_CHG3 },  // Load Byte
  { "ld.h",       CF_USE1|CF_USE2|CF_CHG3 },  // Load Halfword
  { "ld.w",       CF_USE1|CF_USE2|CF_CHG3 },  // Load Halfword
  { "st.b",       CF_USE1|CF_CHG2|CF_USE3 },  // Load Halfword
  { "st.h",       CF_USE1|CF_CHG2|CF_USE3 },  // Load Halfword
  { "st.w",       CF_USE1|CF_CHG2|CF_USE3 },  // Load Halfword
  { "bild",       CF_USE1|CF_CHG2         },  // Load Halfword
  { "bist",       CF_CHG1|CF_USE2         },  // Load Halfword
  { "bdld",       CF_USE1|CF_CHG2         },  // Load Halfword
  { "bdst",       CF_CHG1|CF_USE2         },  // Load Halfword
  /* input/output */
  { "in.b",       CF_USE1|CF_USE2|CF_CHG3 },  // Load Byte
  { "in.h",       CF_USE1|CF_USE2|CF_CHG3 },  // Load Halfword
  { "in.w",       CF_USE1|CF_USE2|CF_CHG3 },  // Load Halfword
  { "out.b",      CF_USE1|CF_CHG2|CF_USE3 },  // Load Halfword
  { "out.h",      CF_USE1|CF_CHG2|CF_USE3 },  // Load Halfword
  { "out.w",      CF_USE1|CF_CHG2|CF_USE3 },  // Load Halfword
 /* arithmetic */
 { "mov", CF_USE1|CF_CHG2},
  { "movhi", CF_USE1|CF_USE2|CF_CHG3 },
  { "add", CF_USE1|CF_USE2|CF_CHG2},
  { "addi", CF_USE1|CF_USE2|CF_CHG3 },
  { "movea", CF_USE1|CF_USE2|CF_CHG3 },
  { "sub", CF_USE1|CF_USE2|CF_CHG2},
  { "mul", CF_USE1|CF_USE2|CF_CHG2},
  { "mulu", CF_USE1|CF_USE2|CF_CHG2},
  { "div", CF_USE1|CF_USE2|CF_CHG2},
  { "divu", CF_USE1|CF_USE2|CF_CHG2},
  { "cmp", CF_USE1|CF_USE2},
  { "setf", CF_USE1|CF_CHG2},
  { "min3", CF_USE1|CF_USE2|CF_CHG3 },
  { "max3", CF_USE1|CF_USE2|CF_CHG3 },
  { "mul3", CF_USE1|CF_USE2|CF_CHG3 },
  { "mac3", CF_USE1|CF_USE2|CF_USE3|CF_CHG3 },
  { "muli", CF_USE1|CF_USE2|CF_CHG3 },
  { "maci", CF_USE1|CF_USE2|CF_USE3|CF_CHG3 },
  { "mult3", CF_USE1|CF_USE2|CF_CHG3 },
  { "mact3", CF_USE1|CF_USE2|CF_USE3|CF_CHG3 },
  { "satadd3", CF_USE1|CF_USE2|CF_CHG3 },
  { "satsub3", CF_USE1|CF_USE2|CF_CHG3 },
  { "or", CF_USE1|CF_USE2|CF_CHG2},
  { "ori", CF_USE1|CF_USE2|CF_CHG3 },
  { "and", CF_USE1|CF_USE2|CF_CHG2},
  { "andi", CF_USE1|CF_USE2|CF_CHG3 },
  { "xor", CF_USE1|CF_USE2|CF_CHG2},
  { "xori", CF_USE1|CF_USE2|CF_CHG3 },
  { "not", CF_USE1|CF_CHG2 },
  { "shl", CF_USE1|CF_USE2|CF_CHG2|CF_SHFT},
  { "shr", CF_USE1|CF_USE2|CF_CHG2|CF_SHFT},
  { "sar", CF_USE1|CF_USE2|CF_CHG2|CF_SHFT},
  { "shld3", CF_USE1|CF_USE2|CF_USE3|CF_CHG3|CF_SHFT},
  { "shrd3", CF_USE1|CF_USE2|CF_USE3|CF_CHG3|CF_SHFT},
  { "jmp", CF_USE1|CF_STOP|CF_JUMP },
  { "jr", CF_USE1|CF_STOP|CF_JUMP },
  { "jal", CF_USE1|CF_CALL|CF_JUMP },
  { "abgt", CF_USE1|CF_JUMP},
  { "bgt", CF_USE1|CF_JUMP},
  { "abge", CF_USE1|CF_JUMP},
  { "bge", CF_USE1|CF_JUMP},
  { "ablt", CF_USE1|CF_JUMP },
  { "blt", CF_USE1|CF_JUMP },
  { "able", CF_USE1|CF_JUMP },
  { "ble", CF_USE1|CF_JUMP },
  { "abh", CF_USE1|CF_JUMP },
  { "bh", CF_USE1|CF_JUMP },
  { "abnl", CF_USE1|CF_JUMP },
  { "bnl", CF_USE1|CF_JUMP },
  { "abl", CF_USE1|CF_JUMP },
  { "bl", CF_USE1|CF_JUMP },
  { "abnh", CF_USE1|CF_JUMP },
  { "bnh", CF_USE1|CF_JUMP },
  { "abe", CF_USE1|CF_JUMP },
  { "be", CF_USE1|CF_JUMP },
  { "abne", CF_USE1|CF_JUMP },
  { "bne", CF_USE1|CF_JUMP },
  { "abv", CF_USE1|CF_JUMP },
  { "bv", CF_USE1|CF_JUMP },
  { "abnv", CF_USE1|CF_JUMP },
  { "bnv", CF_USE1|CF_JUMP },
  { "abn", CF_USE1|CF_JUMP },
  { "bn", CF_USE1|CF_JUMP },
  { "abp", CF_USE1|CF_JUMP },
  { "bp", CF_USE1|CF_JUMP },
  { "abc", CF_USE1|CF_JUMP },
  { "bc", CF_USE1|CF_JUMP },
  { "abnc", CF_USE1|CF_JUMP },
  { "bnc", CF_USE1|CF_JUMP },
  { "abz", CF_USE1|CF_JUMP },
  { "bz", CF_USE1|CF_JUMP },
  { "abnz", CF_USE1|CF_JUMP },
  { "bnz", CF_USE1|CF_JUMP },
  { "abr", CF_USE1|CF_JUMP },
  { "br", CF_USE1|CF_JUMP },
  { "nop", 0 },
  { "ldsr", CF_CHG1|CF_USE2 },
  { "stsr", CF_CHG2|CF_USE1 },
  { "trap", 0 },
  { "reti", CF_STOP },
  { "caxi", 0 },
  { "halt", CF_STOP },
  { "brkret", CF_STOP },
  { "ei", 0 },
  { "di", 0 },
  { "stby", CF_STOP }
};

#ifdef __BORLANDC__
#if sizeof(Instructions)/sizeof(Instructions[0]) != V830_last
#error          No match:  sizeof(InstrNames) !!!
#endif
#endif
